/**
 * CSEE 4840 Embedded System Design
 * Jaycam
 * Yaniv Schiller - yhs2001@columbia.edu
 * Avrum Tilman - amt77@columbia.edu
 * Josh Weinberg - jmw211@columbia.edu
 */

#include "xbasic_types.h"
#include "xio.h"
#include "jaycam.h"

#define W 640
#define H 480
#define VGA_START 0x00800000
#define BRAM_START 0x01800000

// Register addresses for SAA7114H configuration
unsigned char registers [] = {
  
  // Video decoder "generic" registers
  0x01, 0x08, // Recommended setting
  0x02, 0xE4, // Analog input control 1 and input selection Svideo:E8 Analog:E4
  0x03, 0x10, // Analog input control 2
  0x04, 0x90, // Analog input control 3
  0x05, 0x90, // Analog input control 4
  0x06, 0xEB, // Horizontal Sync Start (delay)
  0x07, 0xE0, // Horizontal Sync Stop (delay)
  0x08, 0x59, // Sync control
  0x09, 0x40, // Luminance control Svideo:C0 Analog 40
  0x0A, 0x80,
  0x0B, 0x44,
  0x0C, 0x40,
  0x0D, 0x00,
  0x0E, 0x89,
  0x0F, 0x2A, // Chrominance gain
  0x10, 0x0E, // Chrominance control
  0x11, 0x00,
  0x12, 0x46, // RT signal control
  0x13, 0x00,
  0x14, 0x00,
  0x15, 0x11,
  0x16, 0xFE,
  0x17, 0x40,

  0x18, 0x40,
  0x19, 0x80,
  0x1A, 0x00,
  0x1B, 0x00,
  0x1C, 0x00,
  0x1D, 0x00,
  0x1E, 0x00,
  0x30, 0x08, // Audio clock stuff
  0x31, 0x08, // Audio clock stuff
  0x32, 0x02,

  // Following 1 is added
  0x33, 0x00,

  0x34, 0xCD,
  0x35, 0xCC,
  0x36, 0x3A,

  // Following 1 is added
  0x37, 0x00,

  0x38, 0x03,
  0x39, 0x10,
  0x3A, 0x00,


  // Following 5 are added
  0x3B, 0x00,
  0x3C, 0x00,
  0x3D, 0x00,
  0x3E, 0x00,
  0x3F, 0x00,

  0x40, 0x40,
  0x41, 0xFF,
  0x42, 0xFF,
  0x43, 0xFF,
  0x44, 0xFF,
  0x45, 0xFF, 
  0x46, 0xFF, 
  0x47, 0xFF, 
  0x48, 0xFF,
  0x49, 0xFF,
  0x4A, 0xFF,
  0x4B, 0xFF,
  0x4C, 0xFF,
  0x4D, 0xFF,
  0x4E, 0xFF,
  0x4F, 0xFF,
  0x50, 0xFF,
  0x51, 0xFF,
  0x52, 0xFF,
  0x53, 0xFF,
  0x54, 0xFF,
  0x55, 0xFF,
  0x56, 0xFF,
  0x57, 0xFF,

  0x58, 0x40,
  0x59, 0x47,
  0x5A, 0x06,
  0x5B, 0x03,


  // Following 1 is added
  0x5C, 0x00,

  0x5D, 0x3E,
  0x5E, 0x00,

  // Following 1 is added
  0x5F, 0x00,


  0x80, 0x30,
  0x83, 0x01,
  0x84, 0xA0,
  0x85, 0x10,
  0x86, 0x45,
  0x87, 0x01,
  0x88, 0xF0,

  // Task A Registers
  0x90, 0x00,
  0x91, 0x08,
  0x92, 0x10,
  0x93, 0xC0,
  0x94, 0x10,
  0x95, 0x00,
  0x96, 0xD0,
  0x97, 0x02,
  0x98, 0x0A,
  0x99, 0x00,
  0x9A, 0xF2,
  0x9B, 0x00,
  0x9C, 0xD0, // Horizontal output window size upper bits \ 0xD002 = 720
  0x9D, 0x02, // Horizontal output window size lower bits /          by
  0x9E, 0xF0, // Vertical output window size upper bits  \  0xF000 = 240
  0x9F, 0x00, // Vertical output window size lower bits  /
  0xA0, 0x01,
  0xA1, 0x00,
  0xA2, 0x00,
  0xA4, 0x80,
  0xA5, 0x40,
  0xA6, 0x40,
  0xA8, 0x00,
  0xA9, 0x04,
  0xAA, 0x00,
  0xAC, 0x00,
  0xAD, 0x02,
  0xAE, 0x00,
  0xB0, 0x00,
  0xB1, 0x04,
  0xB2, 0x00,
  0xB3, 0x04,
  0xB4, 0x00,
  0xB8, 0x00,

  // Following 3 were added
  0xB9, 0x00,
  0xBA, 0x00,
  0xBB, 0x00,

  0xBC, 0x00,

  // Following 3 were added
  0xBD, 0x00,
  0xBE, 0x00,
  0xBF, 0x00,
 

  // Task B Registers
  0xC0, 0x00,
  0xC1, 0x08,
  0xC2, 0x10,
  0xC3,	0xC0, 
  0xC4, 0x10,
  0xC5, 0x00,
  0xC6, 0xD0,
  0xC7, 0x02,
  0xC8, 0x0A,
  0xC9, 0x00,
  0xCA, 0xF2,
  0xCB, 0x00,
  0xCC, 0xD0,
  0xCD,	0x02,
  0xCE, 0xF0,
  0xCF, 0x00,
  0xD0, 0x01,
  0xD1, 0x00,
  0xD2, 0x00,
  0xD4, 0x80,
  0xD5, 0x40,
  0xD6, 0x40,
  0xD8, 0x00,
  0xD9,	0x04,
  0xDA, 0x00,
  0xDC, 0x00,
  0xDD, 0x02,
  0xDE, 0x00,
  0xE0, 0x00,
  0xE1, 0x04,
  0xE2, 0x00,
  0xE3, 0x04,
  0xE4, 0x00,
  0xE8,	0x00,
  0xE9, 0x00,
  0xEA, 0x00,
  0xEB, 0x00,
  0xEC,	0x00,
  0xED, 0x00,
  0xEE, 0x00,
  0xEF, 0x00,
				
  // Reset sequence.
  0x88, 0xD8,
  0x88, 0xF8,
  0xFF, 0xFF,};

int w = 0xFF;

void i2c_delay()
{
  int i;
  for (i = 0; i < 10000; i++); 
}

// Write "level" to SCL
void SCLw(int level)
{
  if (level == 0)
    w &= 0xDF;
  else
    w |= 0x2F;

  // Assert the clock on SCL according to level
  XIo_Out8(0xFEFF0200, w);
  i2c_delay();   
}

// Write "level" to SDA
void SDAw(int level)
{
  if (level == 0)
    w &= 0x7F;
  else
    w |= 0x8F;

  // Assert the clock on SDA according to level
  XIo_Out8(0xFEFF0200, w); 
  i2c_delay();  
}

// Read from SDA
int SDAr()
{
  int MSB = XIo_In8(0xFEFF0200);

  MSB = MSB >> 7;
  MSB &= 1;

  i2c_delay();
  return MSB;
}

// Tristate for SDA
void SDAt(int rnw)
{
  if (rnw == 0)
    w &= 0xBF;
  else
    w |= 0x4F;

  // Assert the clock on SDA according to level
  XIo_Out8(0xFEFF0200, w);
  i2c_delay();
}

// Tristate for SCL //
void SCLt(int rnw)
{
  if (rnw == 0)
    w &= 0xEF;
  else
    w |= 0x1F;

  // Assert the clock on SDA according to level
  XIo_Out8(0xFEFF0200, w);
  i2c_delay();
}

// Send the start sequence
void send_start( void )
{
  SCLt(0);
  SDAt(0);
  SCLw(1);
  SDAw(0);
  SCLw(0);
}

// Send the restart sequence
// Needed for read register
// This function must be entered with SDA High
void re_start( void )
{
  SCLw(1);
  SDAw(0);
  SCLw(0);
}


// Send stop sequence
void send_stop( void )
{
  SCLw(0);
  SDAw(0);
  SCLw(1);
  SDAw(1); // Should leave with both lines high to indicate finish
}

// Check acknowledge
int check_ack(void)
{
  int theresult;
  SDAt(1);
  SCLw(1);
  theresult=SDAr();
  SCLw(0);
  SDAw(1); // Set the output before it becomes active to eliminate spike
  SDAt(0);
  return theresult;
}

// Send one bit
void send_bit(int x)
{
  x = x & 1; 
  SDAw(x);
  SCLw(1);
  SCLw(0);
}

// Send an entire bit
void send_byte(int byte)
{
  int i;
  for (i = 7; i >= 0; i--)
    { 
      send_bit(byte >> i);
    }
}

// Read a register from the video decoder
int read_register( int sub_address )
{
  int input = 0;
  int id;

  send_start();

  // Write slave address for SAA7114H is 43H
  send_byte(0x42);

  check_ack();

  // Send the subaddress
  send_byte(sub_address);

  check_ack();

  re_start();

  // Read address
  send_byte(0x43);

  check_ack();
    
  SDAt(1);
  for( id=8 ; id>0 ; id=id-1 )
    {  
      input=input<<1;
      SCLw(1);
      input=input|SDAr();
      SCLw(0);
    }
  // Set the output prior to enable to eliminate spike and 
  // make compatible with Restart
  SDAw(1);
  SDAt(0);
  SCLw(1);
  SCLw(0);
  send_stop();
  return input;
}

// Write a register into the video decoder
void write_register(int sub_address, int data)
{
  int i;

  // Start conditions
  send_start();

  for (i = 0; i < 5; i++)
    i2c_delay();

  // Write slave address for SAA7114H is 42H
  send_byte(0x42);

  check_ack();

  send_byte(sub_address);

  check_ack();

  send_byte(data);

  check_ack();

  send_stop();
}

BYTE packet[PACKET_SIZE];

int main()
{
  Xuint8 bw_1pixel;
  Xuint32 bw_4pixels;
  Xuint32 luma_4pixels;
  Xuint32 current_level;
  Xuint32 vga_addr;

  BYTE my_count=0;

  WORD word;
  WORD *hdr;
  Xuint32 *data;

  int frame_id;
  int line_number = 0;
  int frame_line = 0;
  int start, end;
  int i, j, x;
  int line_section;

  print("\r\n\r\n-------------  Welcome to JayCAM ------------------\r\n\r\n");

  microblaze_enable_icache();

  print("***** Configuring the Video Decoder *****\r\n\r\n");
  // Start the bus  protocol  by  sending
  // a stop handshaking (SDA=1 and SCL=1)
  send_stop();

  i = 0;

  // Configure the video decoder SAA7114H //
  while (registers[i] != 0xFF) {
    write_register (registers[i], registers[i+1]);
    i+=2;
  }

  // Wait and then check how the Philips chip responded to the configuration
  for (i=0; i<100000000;i++);
  if(read_register(0x1F) == 0xA1){
    print("Video decoder configured!\r\n");
  }
  else{
    print("Video decoder not configured!...");
    putnum(read_register(0x1F));
    print("\r\n");
    return;
  }
  
  print("\r\n\r\n***** Ethernet Diagnostic *****\r\n\r\n");
  diag();
  print("\r\n\r\n***** Ethernet Configuration *****\r\n\r\n");
  
  // reset the NIC card
  if(!init_etherne()){
    print("Ethernet NIC not detected");
    return;
  }

  print("\r\n\r\n***** Starting JayCAM *****\r\n\r\n");

  // create non-changing packet header (ethernet, ip, udp headers)
  print("Creating Ethernet, IP, and UDP headers...");
  packet[0]=0x01;
  packet[1]=0x00;
  packet[2]=0x5e;
  packet[3]=0x05;
  packet[4]=0x06;
  packet[5]=0x07;
  packet[6]=0x00;
  packet[6]=0x0D;
  packet[8]=0x60;
  packet[9]=0x7F;
  packet[10]=0xF9;
  packet[11]=0xAF;
  packet[12]=0x08;
  packet[13]=0x00;
  packet[14]=0x45;
  packet[15]=0x00;
  packet[16]=0x00;
  packet[17]=0xBE;
  packet[18]=0x00;
  packet[19]=0x00;
  packet[20]=0x00;
  packet[21]=0x00;
  packet[22]=0x01;
  packet[23]=0x11;
  packet[24]=0xB9;
  packet[25]=0x45;
  packet[26]=0x80;
  packet[27]=0x3B;
  packet[28]=0x95;
  packet[29]=0xA2;
  packet[30]=0xE4;
  packet[31]=0x05;
  packet[32]=0x06;
  packet[33]=0x07;
  packet[34]=0x0B;
  packet[35]=0xE2;
  packet[36]=0x1A;
  packet[37]=0x85;
  packet[38]=0x00;
  packet[39]=0xAA;
  packet[40]=0x00;
  packet[41]=0x00;
  packet[42]=0x00; //extra byte
  print("Done!\r\n");

  print("Writing headers to memory...");
  // access packet contents as 16-bit words
  hdr = (WORD *)&packet;

  outnic(RSAR0, 0); // set DMA starting address
  outnic(RSAR1, TXSTART);

  outnic(ISR, 0xFF); // clear ISR

  outnic(RBCR0, 42); // set Remote DMA Byte Count
  outnic(RBCR1, 0);

  outnic(CMDR, 0x12); // start the DMA write

  // change order of MS/LS since DMA
  // writes LS byte in 15-8, and MS byte in 7-0
  for(i=0; i<21; i++){
    word = (hdr[i]<<8)|(hdr[i]>>8);
    outnic(DATAPORT, word);
  }

  if(innic(ISR)&0x40){
    print("Done!\r\n");
  }
  else{
    print("Error! DMA did not finish. Restart the board\r\n\r\n");    
    return;
  }

  // set 32-bit data pointer for writing to the packet
  // 21 words => 42 bytes of header information
  data = (Xuint32 *)(hdr+21);

  // Wait and then check how the Philips chip responded to the configuration
  for (i=0; i<100000000;i++);
  if(read_register(0x1F) == 0xA1){
    print("Starting video capture...Sending data!\r\n");
  }
  else{
    print("Video failed to start!\r\n\r\n");
    return;
  }

  // start grabbing data
  while (1)
  {
    frame_line = 0;
    line_number = 0;

    // Wait for the vertical synchronism
    while (!(XIo_In32(0x01802FFC)));

    // each iteration corresponds to a full line
    while (1)
    {
      // only use 1 half of the frames
      // (discard interlaced frame)
      if(!XIo_In32(0x018008FC))
	continue;
      
      // save only one-half of the lines
      frame_line = frame_line + 1;
      if(frame_line&0x1)
	line_number++;
       
      packet[43] = (BYTE)(line_number&0xFF);
      data = (Xuint32 *)(&packet[44]);
      
      // This variable indicates how much of
      // the current line has  been  already
      // written into the block RAMs
      current_level = 0x0001;
      if(!(frame_line&0x1)) {
      for (line_section = 0; line_section < 4; line_section++, data+=10)
      {
	// Wait for the current line to be  25%, 50%, 75%, and 100% filled.
	// The loop then executes 4 times
	while (!(XIo_In32(0x01803FFC) & current_level))
	{
	  // reset if a new frame is detected
	  if (!XIo_In32(0x01802FFC))
	    break;
	}

	if (current_level == 0x01) {
	  start = 0;
	  end = 160;
	}
	else if (current_level == 0x02) {
	  start = 160;
	  end = 320;
	}
	else if (current_level == 0x04) {
	  start = 320;
	  end = 480;
	}
	else if (current_level == 0x08) {
	  start = 480;
	  end = 640;
	}

	// reset if a new frame is detected
	if (!XIo_In32(0x01802FFC))
	  break;

	write_video(start, end, line_number, data); // put data in the packet
	current_level = current_level << 1;

	// reset if a new frame is detected
	if (!XIo_In32(0x01802FFC))
	  break;
      } // end of line for loop
      }
      // send every other line
      if(frame_line&0x1)
	send((WORD *)&packet[42], (TXSTART<<8)+42, 162, 204);

      // reset if a new frame is detected
      if (!XIo_In32(0x01802FFC))
	  break;
    } // end of line while loop
  } // end of frame while loop

  print("\r\nExiting JayCAM\r\n\r\n");
  return 0;
}


void write_video(int start, int end, int line, Xuint32 *data)
{
  int nPixs;
  Xuint32 luma_4pixels;
  Xuint32 bram_addr;

  nPixs = (end - start);
  bram_addr = BRAM_START + (start>>2);
  while (nPixs > 0)
  {
    luma_4pixels = XIo_In32(bram_addr+0);
    *data = luma_4pixels;
    data++;

    luma_4pixels = XIo_In32(bram_addr+4);
    *data = luma_4pixels;
    data++;

    bram_addr += 8;
    nPixs -= 32;

    // reset if a new frame is detected
    if (!XIo_In32(0x01802FFC))
      break;
  }
}

// diagnostic that tests nic functionality
void diag(){
  int i;
  
  print("         Command Register Page Switching Test\r\n");
  print("Switching to page 1\r\n");
  outnic(CMDR, 0x61);

  print("Writing to and reading from 0x0D (value should be 0x4e): ");
  outnic(0x0D*2,0x4e);
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 0\r\n");
  outnic(CMDR, 0x21);

  print("Reading from reg offset 0x0D: ");
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 1\r\n");
  outnic(CMDR, 0x61);
  print("Reading from reg offset 0x0D (should be 0x4e): ");
  putnum(innic(0x0D*2));
  print("\r\n\r\n");

  print("         Default Value Test\r\n");
  print("Switching to page 0\r\n");
  outnic(CMDR,21);
  print("Reading from 0x16 (value should be 0x15): ");
  putnum(innic(0x16*2));
  print("\r\n");
  print("Reading from 0x12 (value should be 0x0c): ");
  putnum(innic(0x12*2));
  print("\r\n");
  print("Reading from 0x13 (value should be 0x12): ");
  putnum(innic(0x13*2));
  print("\r\n");
}

int init_etherne()
{
  outnic(NE_RESET, innic(NE_RESET));  // Do reset
  delay(2);
  if ((innic(ISR) & 0x80) == 0)       // Report if failed
  {
    print("  Ethernet card failed to reset!\r\n");
    return 0;
  }
  else
  {
    print("Ethernet card reset successful!\r\n");
    resetnic(); // Reset Ethernet card,
    print("Ethernet card intialization complete!\r\n");
  }
  return 1;
}

void resetnic(){
  int i, count;

  outnic(CMDR, 0x21);  // Abort and DMA and stop the NIC
  delay(10);

  outnic(DCR, DCRVAL);   // Set word-wide access

  outnic(RBCR0, 0x00); // Clear the count regs
  outnic(RBCR1, 0x00);

  outnic(IMR, 0x00);   // Mask completion irq
  outnic(ISR, 0xFF);   // clear interrupt status register

  outnic(RCR, 0x20);   // 0x20  Set to monitor mode
  outnic(TCR, 0x02);  // put nic in normal operation

  // Set Rx start, Rx stop, Boundary and TX start regs
  outnic(PSTART, RXSTART);
  outnic(PSTOP, RXSTOP);
  outnic(BNRY, (BYTE)(RXSTOP-1));
  outnic(TPSR, TXSTART);
  
  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(IMR, 0x00);   // Mask completion irq

  // put nic in start mode
  outnic(CMDR, 0x22); // start nic
  outnic(TCR, 0x00);  // put nic in normal operation
}

void delay(int mult){
  int i;
  int delay = 1000000*mult;
  for(i=0; i<delay; i++);
}
